/*
 * SPCA4ZC based usb camera driver (currently spports zc301p+hv7131r) 
 * SPCA4ZC version by Kim Sungjo <kauri.kim@gmail.com>
 * 
 * Based on :
 * SPCA5XX version by Michel Xhaard <mxhaard@users.sourceforge.net>
 * ZC030X device driver <http://zc0302.sourceforge.net>
 *
 * This program is under GPL. 
 */

static const char version[] = SPCA5XX_VERSION;

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/init.h>

#include <linux/fs.h>
#include <linux/vmalloc.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/mm.h>
#include <linux/string.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/ctype.h>
#include <linux/pagemap.h>
#include <linux/usb.h>

#include <asm/io.h>
#include <asm/semaphore.h>
#include <asm/page.h>
#include <asm/uaccess.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
/* only on 2.6.x */
#include <linux/jiffies.h>
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
#include <linux/wrapper.h>
#endif
#include <linux/param.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 9)
#include <linux/moduleparam.h>
#endif

#include "spca5xx.h"

#include "spcadecoder.h"
#include "zc3xx.h"
#include "spca_isoc.h"
#include "spca_mem.h"
#include "spca_v4l.h"
#include "spca_mode.h"

#define PROC_NAME_LEN 10	//length of the proc name

/* Hardware auto exposure / whiteness (PC-CAM 600) */
static int autoexpo = 1;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,5)
/* Video device number (-1 is first available) */
static int video_nr = -1;
#endif				/* LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,5) */

/* 0=no debug messages
 * 1=init/detection/unload and other significant messages,
 * 2=some warning messages
 * 3=config/control function calls
 * 4=most function calls and data parsing messages
 * 5=highly repetitive mesgs
 * NOTE: This should be changed to 0, 1, or 2 for production kernels
 */
int debug = 3;

/* Force image to be read in RGB instead of BGR. This option allow
 * programs that expect RGB data (e.g. gqcam) to work with this driver. */
static int force_rgb = 0;
static int gamma = 3;
static int OffRed = 0;
static int OffBlue = 0;
static int OffGreen = 0;
static int GRed = 256;
static int GBlue = 256;
static int GGreen = 256;

static int usbgrabber = 0;

#ifdef SPCA50X_ENABLE_COMPRESSION
/* Enable compression. This is for experimentation only; compressed images
 * still cannot be decoded yet. */
static int compress = 0;
#endif				/* SPCA50X_ENABLE_COMPRESSION */

#ifdef SPCA5XX_ENABLE_REGISTERPLAY
static int RegAddress = 0;
static int RegValue = 0;
static int RegStrobe = 0;
#endif				/* SPCA5XX_ENABLE_REGISTERPLAY */

/* Initial brightness & contrast (for debug purposes) */
static int bright = 0x80;
static int contrast = 0x60;

/* Parameter that enables you to set the minimal suitable bpp */
static int min_bpp = 0;

/* Parameter defines the average luminance that should be kept */
static int lum_level = 0x2d;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 9)

module_param(autoexpo, int, 0644);
module_param(debug, int, 0644);
module_param(force_rgb, int, 0644);
module_param(gamma, int, 0644);
module_param(OffRed, int, 0644);
module_param(OffBlue, int, 0644);
module_param(OffGreen, int, 0644);
module_param(GRed, int, 0644);
module_param(GBlue, int, 0644);
module_param(GGreen, int, 0644);

#ifdef SPCA50X_ENABLE_COMPRESSION
module_param(compress, int, 0644);
#endif				/* SPCA50X_ENABLE_COMPRESSION */
module_param(bright, int, 0444);
module_param(contrast, int, 0444);

module_param(min_bpp, int, 0444);
module_param(lum_level, int, 0444);
module_param(usbgrabber, int, 0444);
#ifdef SPCA5XX_ENABLE_REGISTERPLAY
module_param(RegAddress, int, 0644);
module_param(RegValue, int, 0644);
module_param(RegStrobe, int, 0644);
#endif				/* SPCA5XX_ENABLE_REGISTERPLAY */

#else				/* LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 9) */

MODULE_PARM(autoexpo, "i");
MODULE_PARM(debug, "i");

MODULE_PARM(force_rgb, "i");
MODULE_PARM(gamma, "i");
MODULE_PARM(OffRed, "i");
MODULE_PARM(OffBlue, "i");
MODULE_PARM(OffGreen, "i");
MODULE_PARM(GRed, "i");
MODULE_PARM(GBlue, "i");
MODULE_PARM(GGreen, "i");

#ifdef SPCA50X_ENABLE_COMPRESSION
MODULE_PARM(compress, "i");
#endif				/* SPCA50X_ENABLE_COMPRESSION */
MODULE_PARM(bright, "i");
MODULE_PARM(contrast, "i");

MODULE_PARM(min_bpp, "i");
MODULE_PARM(lum_level, "0-255i");
MODULE_PARM(usbgrabber, "i");
#endif
/***************/

MODULE_PARM_DESC(autoexpo,
		 "Enable/Disable auto exposure (default=1: enabled) (PC-CAM 600/Zc03xx/spca561a/Etoms Only !!!)");
MODULE_PARM_DESC(debug,
		 "Debug level: 0=none, 1=init/detection, 2=warning, 3=config/control, 4=function call, 5=max");

MODULE_PARM_DESC(force_rgb, "Read RGB instead of BGR");
MODULE_PARM_DESC(gamma, "gamma setting range 0 to 7 3-> gamma=1");
MODULE_PARM_DESC(OffRed, "OffRed setting range -128 to 128");
MODULE_PARM_DESC(OffBlue, "OffBlue setting range -128 to 128");
MODULE_PARM_DESC(OffGreen, "OffGreen setting range -128 to 128");
MODULE_PARM_DESC(GRed, "Gain Red setting range 0 to 512 /256 ");
MODULE_PARM_DESC(GBlue, "Gain Blue setting range 0 to 512 /256 ");
MODULE_PARM_DESC(GGreen, "Gain Green setting range 0 to 512 /256 ");

#ifdef SPCA50X_ENABLE_COMPRESSION
MODULE_PARM_DESC(compress, "Turn on/off compression (not functional yet)");
#endif				/* SPCA50X_ENABLE_COMPRESSION */
MODULE_PARM_DESC(bright,
		 "Initial brightness factor (0-255) not know by all webcams !!");
MODULE_PARM_DESC(contrast,
		 "Initial contrast factor (0-255) not know by all webcams !!");
MODULE_PARM_DESC(min_bpp,
		 "The minimal color depth that may be set (default 0)");
MODULE_PARM_DESC(lum_level,
		 "Luminance level for brightness autoadjustment (default 32)");
MODULE_PARM_DESC(usbgrabber, "Is a usb grabber 0x0733:0x0430 ? (default 1) ");

#ifdef SPCA5XX_ENABLE_REGISTERPLAY
MODULE_PARM_DESC(RegAddress, "Register Address of PAC207");
MODULE_PARM_DESC(RegValue, "Register Value for PAC207");
MODULE_PARM_DESC(RegStrobe,
		 "Strobe to read or write a register 1=write, 2=read");
#endif				/* SPCA5XX_ENABLE_REGISTERPLAY */

/****************/
MODULE_AUTHOR
    ("Michel Xhaard <mxhaard@users.sourceforge.net> based on spca50x driver by Joel Crisp <cydergoth@users.sourceforge.net>,ov511 driver by Mark McClelland <mwm@i.am>");
MODULE_DESCRIPTION("SPCA5XX USB Camera Driver");
MODULE_LICENSE("GPL");

//static int spca50x_move_data(struct usb_spca50x *spca50x, struct urb *urb);

static struct usb_driver spca5xx_driver;

#ifndef max
static inline int max(int a, int b)
{
	return (a > b) ? a : b;
}
#endif				/* max */

/**********************************************************************
 * List of known SPCA50X-based cameras
 **********************************************************************/

/* Camera type jpeg yuvy yyuv yuyv grey gbrg*/
static struct palette_list Plist[] = {
	{JPEG, "JPEG"},
	{JPGH, "JPEG"},
	{JPGC, "JPEG"},
	{JPGS, "JPEG"},
	{JPGM, "JPEG"},
	{YUVY, "YUVY"},
	{YYUV, "YYUV"},
	{YUYV, "YUYV"},
	{GREY, "GREY"},
	{GBRG, "GBRG"},
	{SN9C, "SN9C"},
	{GBGR, "GBGR"},
	{S561, "S561"},
	{PGBRG, "GBRG"},
	{-1, NULL}
};

static struct bridge_list Blist[] = {
	{BRIDGE_ZC3XX, "ZC301-2"},
	{-1, NULL}
};

enum {
	UnknownCamera = 0,	// 0
	Vimicro
};
static struct cam_list clist[] = {
	{UnknownCamera, "Unknown"},
	{Vimicro, "Z-star Vimicro zc0301p"},
	{-1, NULL}
};

static __devinitdata struct usb_device_id device_table[] = {
	
	{USB_DEVICE(0x0ac8, 0x301b)},	/* Asam Vimicro */	/* our:kauri */
	{}			/* Terminating entry */
};

MODULE_DEVICE_TABLE(usb, device_table);

#define GET_EXT_MODES(bridge) (spca50x_ext_modes)

/* code is used for harware low nibble and software hight nibble */
static __u16 spca50x_ext_modes[][6] = {
	/* x , y , Code, Value (6), Value (7), pipe */
	{640, 480, 0, 0x10, 0x10, 1023},	/* Tested spca505b+VGA sensor */
	{352, 288, 1, 0x1a, 0x1a, 1023},	/* Tested */
	{320, 240, 2, 0x1c, 0x1d, 896},	/* Tested 20:01:2004 */
	{176, 144, 4, 0x34, 0x34, 512},	/* Tested */
	{160, 120, 5, 0x40, 0x40, 384},	/* Tested */
	{0, 0, 0, 0, 0}
};

#ifdef CONFIG_PROC_FS
# undef CONFIG_PROC_FS
#endif

#ifdef CONFIG_PROC_FS
/* Not sure what we should do with this. I think it is V4L level 2 stuff */
/* Currently only use RGB24 */
static struct palette_list plist[] = {
	{VIDEO_PALETTE_GREY, "GREY"},
	{VIDEO_PALETTE_HI240, "HI240"},
	{VIDEO_PALETTE_RGB565, "RGB565"},
	{VIDEO_PALETTE_RGB24, "RGB24"},
	{VIDEO_PALETTE_RGB32, "RGB32"},
	{VIDEO_PALETTE_RGB555, "RGB555"},
	{VIDEO_PALETTE_YUV422, "YUV422"},
	{VIDEO_PALETTE_YUYV, "YUYV"},
	{VIDEO_PALETTE_UYVY, "UYVY"},
	{VIDEO_PALETTE_YUV420, "YUV420"},
	{VIDEO_PALETTE_YUV411, "YUV411"},
	{VIDEO_PALETTE_RAW, "RAW"},
	{VIDEO_PALETTE_YUV422P, "YUV422P"},
	{VIDEO_PALETTE_YUV411P, "YUV411P"},
	{VIDEO_PALETTE_YUV420P, "YUV420P"},
	{VIDEO_PALETTE_YUV410P, "YUV410P"},
	{VIDEO_PALETTE_RAW_JPEG, "RJPG"},
	{VIDEO_PALETTE_JPEG, "JPEG"},
	{-1, NULL}
};
#endif				/* CONFIG_PROC_FS */

/* function for the tasklet */

#if 0 /* don't need */
static inline int
spca50x_smallest_mode_index(__u16 ext_modes[][6],
			    int *cols, int *rows, int *pipe)
{
	int i;
	int width = INT_MAX;
	int height = INT_MAX;
	int intpipe = 1023;
	int index = -EINVAL;

	/* search through the mode list until we run out of modes */
	/* and take the smallest hardware mode and pipe size */
	for (i = 0; ext_modes[i][0]; i++) {
		if (ext_modes[i][0] < width || ext_modes[i][1] < height) {
			if (!(ext_modes[i][2] & 0xF0)) {
				width = ext_modes[i][0];
				height = ext_modes[i][1];
				intpipe = ext_modes[i][5];
				index = i;
			}
		}
	}

	/* return the index of the smallest mode */
	if (index != -EINVAL) {
		*pipe = intpipe;
		if (cols != NULL)
			*cols = width;
		if (rows != NULL)
			*rows = height;
	}
	return i;
}
#endif


#if 0
static inline void spca5xx_chgAuto(struct usb_spca50x *spca50x, __u8 autoval)
{
	spca50x->autoexpo = (int)autoval;
	switch (spca50x->bridge) {
	case BRIDGE_ZC3XX:
		zc3xx_setAutobright(spca50x);
		break;
	default:
		break;
	}
}
static inline void spca5xx_chgQtable(struct usb_spca50x *spca50x, __u8 qtable)
{
/* one frame maybe corrupted  wait for the result */

	switch (spca50x->bridge) {
	case BRIDGE_ZC3XX:
		spca50x->qindex = (int)qtable;
		zc3xx_setquality(spca50x);
		init_jpeg_decoder(spca50x);
		break;
	default:
		break;
	}

}
static inline void spca5xx_chgDtimes(struct usb_spca50x *spca50x, __u16 dtimes)
{
	unsigned long flags;
	spin_lock_irqsave(&spca50x->v4l_lock, flags);
	spca50x->dtimes = (unsigned int)dtimes;
	spin_unlock_irqrestore(&spca50x->v4l_lock, flags);
}
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,22)
static int spca5xx_open(struct inode *inode, struct file *file)
{
	struct video_device *vdev = video_devdata(file);
#else
static int spca5xx_open(struct video_device *vdev, int flags)
{
#endif
	struct usb_spca50x *spca50x = video_get_drvdata(vdev);

	int err;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	MOD_INC_USE_COUNT;
#endif
	PDEBUG(2, "opening");

	down(&spca50x->lock);
	/* sanity check disconnect, in use, no memory available */
	err = -ENODEV;
	if (!spca50x->present)
		goto out;
	err = -EBUSY;
	if (spca50x->user)
		goto out;
	err = -ENOMEM;
	if (spca50x_alloc(spca50x))
		goto out;
	/* initialize sensor and decoding */
	err = spca50x_init_source(spca50x);
	if (err != 0) {
		PDEBUG(0, "DEALLOC error on spca50x_init_source\n");
		up(&spca50x->lock);
		spca5xx_dealloc(spca50x);
		goto out2;
	}
	/* jpeg header quality */
	init_jpeg_decoder(spca50x);
	//spca5xx_initDecoder(spca50x);
	spca5xx_setFrameDecoder(spca50x);

	spca50x->user++;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,22)
	file->private_data = vdev;
#endif
	/* init usb isoc transfer */
	err = spca50x_init_isoc(spca50x);
	if (err) {
		PDEBUG(0, " DEALLOC error on init_Isoc\n");
		spca50x->user--;
		spca5xx_kill_isoc(spca50x);
		up(&spca50x->lock);
		spca5xx_dealloc(spca50x);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,22)
		file->private_data = NULL;
#endif
		goto out2;
	}

	/* Now, let's get brightness from the camera */
	spca50x->brightness = zc3xx_getbrightness(spca50x) << 8;
	spca50x->whiteness = 0;

      out:
	up(&spca50x->lock);
      out2:
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	if (err)
		MOD_DEC_USE_COUNT;
#endif

	if (err) {
		PDEBUG(2, "Open failed");
	} else {
		PDEBUG(2, "Open done");
	}

	return err;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,22)
static int spca5xx_close(struct inode *inode, struct file *file)
{

	struct video_device *vdev = file->private_data;
#else
static void spca5xx_close(struct video_device *vdev)
{
#endif
	struct usb_spca50x *spca50x = video_get_drvdata(vdev);
	int i;
	PDEBUG(2, "spca50x_close");

	down(&spca50x->lock);

	spca50x->user--;
	spca50x->curframe = -1;
	if (spca50x->present) {
		spca50x_stop_isoc(spca50x);
		zc3xx_shutdown(spca50x);
		// spcaCameraShutDown(spca50x);

		for (i = 0; i < SPCA50X_NUMFRAMES; i++) {
			if (waitqueue_active(&spca50x->frame[i].wq))
				wake_up_interruptible(&spca50x->frame[i].wq);
		}
		if (waitqueue_active(&spca50x->wq))
			wake_up_interruptible(&spca50x->wq);

	}
	/* times to dealloc ressource */
	up(&spca50x->lock);
	spca5xx_dealloc(spca50x);
	PDEBUG(2, "Release ressources done");

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	MOD_DEC_USE_COUNT;
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,22)
	file->private_data = NULL;
	return 0;
#endif
}


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,22)
static int
spca5xx_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
		unsigned long arg)
{
	int rc;
#if 0
	struct video_device *vdev = file->private_data;
	struct usb_spca50x *spca50x = video_get_drvdata(vdev);

	/* lock the device here is dramatic for latencies */
	if (down_interruptible(&spca50x->lock))
		return -EINTR;
#endif
	rc = video_usercopy(inode, file, cmd, arg, spca5xx_do_ioctl);
#if 0
	/* see 7 lines up */ */
		up(&spca50x->lock);
#endif
	return rc;
}
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,22)
static ssize_t
spca5xx_read(struct file *file, char *buf, size_t cnt, loff_t * ppos)
{

	struct video_device *dev = file->private_data;
	int noblock = file->f_flags & O_NONBLOCK;
	unsigned long count = cnt;
#else
static long
spca5xx_read(struct video_device *dev, char *buf, unsigned long
	     count, int noblock)
{
#endif
	struct usb_spca50x *spca50x = video_get_drvdata(dev);

	int i;
	int frmx = -1;
	int rc;
	volatile struct spca50x_frame *frame;

	PDEBUG(4, "%ld bytes, noblock=%d", count, noblock);

	if (down_interruptible(&spca50x->lock))
		return -ERESTARTSYS;	//-EINTR;

	if (!dev || !buf) {
		up(&spca50x->lock);
		return -EFAULT;
	}
	if (!spca50x->dev) {
		up(&spca50x->lock);
		return -EIO;
	}
	if (!spca50x->streaming) {
		up(&spca50x->lock);
		return -EIO;
	}

	/* Wait while we're grabbing the image */
	PDEBUG(4, "Waiting for a frame done");
#if (SPCA50X_NUMFRAMES == 4)
	if ((rc = wait_event_interruptible(spca50x->wq,
					   spca50x->frame[0].grabstate ==
					   FRAME_DONE
					   || spca50x->frame[1].grabstate ==
					   FRAME_DONE
					   || spca50x->frame[2].grabstate ==
					   FRAME_DONE
					   || spca50x->frame[3].grabstate ==
					   FRAME_DONE))) {
		up(&spca50x->lock);
		return rc;
	}
#else				// should be two
	if ((rc = wait_event_interruptible(spca50x->wq,
					   spca50x->frame[0].grabstate ==
					   FRAME_DONE
					   || spca50x->frame[1].grabstate ==
					   FRAME_DONE))) {
		up(&spca50x->lock);
		return rc;
	}
#endif

	/* One frame has just been set to DONE. Find it. */
	for (i = 0; i < SPCA50X_NUMFRAMES; i++)
		if (spca50x->frame[i].grabstate == FRAME_DONE)
			frmx = i;

	PDEBUG(4, "Frame number: %d", frmx);
	if (frmx < 0) {
		/* We havent found a frame that is DONE. Damn. Should
		 * not happen. */
		PDEBUG(2, "Couldnt find a frame ready to be read.");
		up(&spca50x->lock);
		return -EFAULT;
	}
	frame = &spca50x->frame[frmx];
	if (count > frame->scanlength)
		count = frame->scanlength;

	if ((i = copy_to_user(buf, frame->data, count))) {
		PDEBUG(2, "Copy failed! %d bytes not copied", i);
		up(&spca50x->lock);
		return -EFAULT;
	}
	/* Release the frame */
	frame->grabstate = FRAME_READY;
	
	up(&spca50x->lock);

	return count;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,22)
static int spca5xx_mmap(struct file *file, struct vm_area_struct *vma)
{

	struct video_device *dev = file->private_data;
	unsigned long start = vma->vm_start;
	unsigned long size = vma->vm_end - vma->vm_start;
#else
static int
spca5xx_mmap(struct video_device *dev, const char *adr, unsigned long size)
{
	unsigned long start = (unsigned long)adr;
#endif
	struct usb_spca50x *spca50x = video_get_drvdata(dev);
	unsigned long page, pos;

	if (spca50x->dev == NULL)
		return -EIO;

	PDEBUG(4, "mmap: %ld (%lX) bytes", size, size);

	if (size >
	    (((SPCA50X_NUMFRAMES * MAX_DATA_SIZE) + PAGE_SIZE -
	      1) & ~(PAGE_SIZE - 1)))
		return -EINVAL;
	if (down_interruptible(&spca50x->lock))
		return -EINTR;

	pos = (unsigned long)spca50x->fbuf;
	while (size > 0) {
		page = kvirt_to_pa(pos);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0) || defined (RH9_REMAP)
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,9)
		if (remap_pfn_range
		    (vma, start, page >> PAGE_SHIFT, PAGE_SIZE, PAGE_SHARED)) {
#else
		if (remap_page_range(vma, start, page, PAGE_SIZE, PAGE_SHARED)) {
#endif				/* KERNEL 2.6.9 */
#else				/* RH9_REMAP */
		if (remap_page_range(start, page, PAGE_SIZE, PAGE_SHARED)) {
#endif				/* RH9_REMAP */
			up(&spca50x->lock);
			return -EAGAIN;
		}
		start += PAGE_SIZE;
		pos += PAGE_SIZE;
		if (size > PAGE_SIZE)
			size -= PAGE_SIZE;
		else
			size = 0;
	}
	up(&spca50x->lock);
	return 0;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION (2,4,22)
static struct file_operations spca5xx_fops = {
	.owner = THIS_MODULE,
	.open = spca5xx_open,
	.release = spca5xx_close,
	.read = spca5xx_read,
	.mmap = spca5xx_mmap,
	.ioctl = spca5xx_ioctl,
	.llseek = no_llseek,
};
static struct video_device spca50x_template = {
	.owner = THIS_MODULE,
	.name = "SPCA5XX USB Camera",
	.type = VID_TYPE_CAPTURE,
	.hardware = VID_HARDWARE_SPCA5XX,
	.fops = &spca5xx_fops,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
	.release = video_device_release,
#endif
	.minor = -1,
};
#else
static struct video_device spca50x_template = {
      name:"SPCA5XX USB Camera",
      type:VID_TYPE_CAPTURE,
      hardware:VID_HARDWARE_SPCA5XX,
      open:spca5xx_open,
      close:spca5xx_close,
      read:spca5xx_read,
      mmap:spca5xx_mmap,
      ioctl:spca5xx_ioctl
};
#endif				//KERNEL VERSION 2,4,22
/****************************************************************************
 *
 * SPCA50X configuration
 *
 ***************************************************************************/

static int spca50x_configure_sensor(struct usb_spca50x *spca50x)
{

	PDEBUG(4, "starting configuration");
	/* Set sensor-specific vars */
	spca50x->maxwidth = 640;
	spca50x->maxheight = 480;
	spca50x->minwidth = 160;
	spca50x->minheight = 120;
	return 0;
}

/* ̽   ʱ  */
static int spca50x_configure(struct usb_spca50x *spca50x)
{
	int i=0;
	int defaultcols =320;
	int defaultrows =240;
	int defaultpipe =896;

#if 0
	i = spca50x_smallest_mode_index(GET_EXT_MODES(spca50x->bridge),
					&defaultcols, &defaultrows,
					&defaultpipe);
#endif
	if (i < 0)
		return i;

	PDEBUG(2, "video_register_device succeeded");
	/* Initialise the camera bridge */
	switch (spca50x->bridge) {
	case BRIDGE_ZC3XX:
		if (zc3xx_config(spca50x) < 0)
			goto error;
		break;
	default:
		{
			err("Unknown bridge type");
			goto error;
		}
	}

	/* data transfer disable */
	spca50x_set_packet_size(spca50x, 0);
	/* Set an initial pipe size; this will be overridden by
	 * spca50x_set_mode(), called indirectly by the open routine.
	 */
	if (spca50x->bridge == BRIDGE_ZC3XX) {
		/* ⺻ (CIF::320*240)  */
		i = spca5xx_getDefaultMode(spca50x);
		if (i < 0)
			return i;
	} else {
		spca50x->pipe_size = defaultpipe;
		spca50x->width = defaultcols;
		spca50x->height = defaultrows;
	}
	spca50x->force_rgb = force_rgb;
	spca50x->min_bpp = min_bpp;
	spca50x->lum_level = lum_level;
	if (spca50x_configure_sensor(spca50x) < 0) {
		err("failed to configure");
		goto error;
	}

	/* spca50x->maxwidth = 640;
	   spca50x->maxheight = 480;
	   spca50x->minwidth = 160;
	   spca50x->minheight = 120; */

	/* configure the frame detector with default parameters */
	spca5xx_setFrameDecoder(spca50x);

	PDEBUG(2, "Spca5xx Configure done !!");
	return 0;

      error:

	return -EBUSY;
}

/************************************************************************************/

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
/****************************************************************************
 *  sysfs
 ***************************************************************************/

static inline struct usb_spca50x *cd_to_spca50x(struct class_device *cd)
{
	struct video_device *vdev = to_video_device(cd);
	return video_get_drvdata(vdev);
}

static ssize_t show_stream_id(struct class_device *cd, char *buf)
{
	struct usb_spca50x *spca50x = cd_to_spca50x(cd);
	return snprintf(buf, 5, "%s\n", Plist[spca50x->cameratype].name);
}

static CLASS_DEVICE_ATTR(stream_id, S_IRUGO, show_stream_id, NULL);

static ssize_t show_model(struct class_device *cd, char *buf)
{
	struct usb_spca50x *spca50x = cd_to_spca50x(cd);
	return snprintf(buf, 32, "%s\n", (spca50x->desc) ?
			clist[spca50x->desc].description : " Unknow ");
}

static CLASS_DEVICE_ATTR(model, S_IRUGO, show_model, NULL);

static ssize_t show_pictsetting(struct class_device *cd, char *buf)
{
	struct usb_spca50x *spca50x = cd_to_spca50x(cd);
	struct pictparam *gcorrect = &spca50x->pictsetting;
	return snprintf(buf, 128,
			"force_rgb=%d, gamma=%d, OffRed=%d, OffBlue=%d, OffGreen=%d, GRed=%d, GBlue=%d, GGreen= %d \n",
			gcorrect->force_rgb, gcorrect->gamma, gcorrect->OffRed,
			gcorrect->OffBlue, gcorrect->OffGreen, gcorrect->GRed,
			gcorrect->GBlue, gcorrect->GGreen);
}

static CLASS_DEVICE_ATTR(pictsetting, S_IRUGO, show_pictsetting, NULL);
/*
static ssize_t store_RGB(struct class_device *cd, char *buf,size_t count)
{
	struct usb_spca50x *spca50x = cd_to_spca50x(cd);
	struct pictparam *gcorrect = &spca50x->pictsetting;
	int ret;
	ret= sscanf(buf,"%d",&gcorrect->force_rgb);
	if (ret != 1) return -EINVAL;
	gcorrect->change = 0x10;
	return strlen(buf);
} 
static CLASS_DEVICE_ATTR(RGB, S_IWUGO, NULL, store_RGB);

static ssize_t show_brightness(struct class_device *cd, char *buf)
{
	struct usb_spca50x *spca50x = cd_to_spca50x(cd);
	unsigned short x;

	if (!spca50x->dev)
		return -ENODEV;
	sensor_get_brightness(spca50x, &x);
	return sprintf(buf, "%d\n", x >> 8);
} 
static CLASS_DEVICE_ATTR(brightness, S_IRUGO, show_brightness, NULL);

static ssize_t show_saturation(struct class_device *cd, char *buf)
{
	struct usb_spca50x *spca50x = cd_to_spca50x(cd);
	unsigned short x;

	if (!spca50x->dev)
		return -ENODEV;
	sensor_get_saturation(spca50x, &x);
	return sprintf(buf, "%d\n", x >> 8);
} 
static CLASS_DEVICE_ATTR(saturation, S_IRUGO, show_saturation, NULL);

static ssize_t show_contrast(struct class_device *cd, char *buf)
{
	struct usb_spca50x *spca50x = cd_to_spca50x(cd);
	unsigned short x;

	if (!spca50x->dev)
		return -ENODEV;
	sensor_get_contrast(spca50x, &x);
	return sprintf(buf, "%d\n", x >> 8);
} 
static CLASS_DEVICE_ATTR(contrast, S_IRUGO, show_contrast, NULL);

static ssize_t show_hue(struct class_device *cd, char *buf)
{
	struct usb_spca50x *spca50x = cd_to_spca50x(cd);
	unsigned short x;

	if (!spca50x->dev)
		return -ENODEV;
	sensor_get_hue(spca50x, &x);
	return sprintf(buf, "%d\n", x >> 8);
} 
static CLASS_DEVICE_ATTR(hue, S_IRUGO, show_hue, NULL);

static ssize_t show_exposure(struct class_device *cd, char *buf)
{
	struct usb_spca50x *spca50x = cd_to_spca50x(cd);
	unsigned char exp;

	if (!spca50x->dev)
		return -ENODEV;
	sensor_get_exposure(spca50x, &exp);
	return sprintf(buf, "%d\n", exp >> 8);
} 
static CLASS_DEVICE_ATTR(exposure, S_IRUGO, show_exposure, NULL);
*/

static void spca50x_create_sysfs(struct video_device *vdev)
{

	video_device_create_file(vdev, &class_device_attr_stream_id);
	video_device_create_file(vdev, &class_device_attr_model);
	video_device_create_file(vdev, &class_device_attr_pictsetting);
/*	
	video_device_create_file(vdev, &class_device_attr_RGB);

	video_device_create_file(vdev, &class_device_attr_brightness);
	video_device_create_file(vdev, &class_device_attr_saturation);
	video_device_create_file(vdev, &class_device_attr_contrast);
	video_device_create_file(vdev, &class_device_attr_hue);
	video_device_create_file(vdev, &class_device_attr_exposure);
*/
}
#endif
/****************************************************************************
 *
 *  USB routines
 *
 ***************************************************************************/
static int spcaDetectCamera(struct usb_spca50x *spca50x)
{
	struct usb_device *dev = spca50x->dev;
	//__u8 fw = 0;
	__u16 vendor;
	__u16 product;
	/* Is it a recognised camera ? */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,11)
	vendor = le16_to_cpu(dev->descriptor.idVendor);
	product = le16_to_cpu(dev->descriptor.idProduct);
#else
	vendor = dev->descriptor.idVendor;
	product = dev->descriptor.idProduct;
#endif
	switch (vendor) {
	case 0x0AC8:		/* Vimicro z-star */
	case 0x15B8:
		switch (product) {
		case 0x301b:	/* Wasam 350r */
		case 0x5002:	/* kauri: our cam */
			spca50x->desc = Vimicro;
			spca50x->bridge = BRIDGE_ZC3XX;
			spca50x->sensor = SENSOR_HV7131R;	//overwrite by the sensor detect routine
			spca50x->header_len = SPCA50X_OFFSET_DATA;
			spca50x->cameratype = JPGH;
			info("USB SPCA5XX camera found. Type Vimicro Zc301P 0x301b");

			break;
		default:
			goto error;
		};
		break;
	default:
		goto error;
	}
	return 0;
      error:
	return -ENODEV;
}

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
static int
spca5xx_probe(struct usb_interface *intf, const struct usb_device_id *id)
#else
static void *spca5xx_probe(struct usb_device *dev, unsigned int ifnum,
			   const struct usb_device_id *id)
#endif
{
	struct usb_interface_descriptor *interface;
	struct usb_spca50x *spca50x;
	int err_probe;
	int i;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
	struct usb_device *dev = interface_to_usbdev(intf);
#endif
	/* We don't handle multi-config cameras */
	if (dev->descriptor.bNumConfigurations != 1)
		goto nodevice;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	/* 2.4.x test Only work on Interface 0 */
	if (ifnum > 0)
		goto nodevice;

	interface = &dev->actconfig->interface[ifnum].altsetting[0];
	/* Since code below may sleep, we use this as a lock */
	MOD_INC_USE_COUNT;
#else
	/* 2.6.x test Only work on Interface 0 */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,6)
	interface = &intf->cur_altsetting->desc;
#else
	interface = &intf->altsetting[0].desc;
#endif

	if (interface->bInterfaceNumber > 0)
		goto nodevice;
#endif
	PDEBUG(0, ">>> spca5xx_probe");

	if ((spca50x = kmalloc(sizeof(struct usb_spca50x), GFP_KERNEL)) == NULL) {
		err("couldn't kmalloc spca50x struct");
		goto error;
	}

	memset(spca50x, 0, sizeof(struct usb_spca50x));

	spca50x->dev = dev;
	spca50x->iface = interface->bInterfaceNumber;
	/* vender, product ID ȮϿ  */
	if ((err_probe = spcaDetectCamera(spca50x)) < 0) {
		err(" Devices not found !! ");
		/* FIXME kfree spca50x and goto nodevice */
		goto error;
	}
	PDEBUG(0, "Camera type %s ", Plist[spca50x->cameratype].name);

	/*  ۿ  ť  */
	for (i = 0; i < SPCA50X_NUMFRAMES; i++)
		init_waitqueue_head(&spca50x->frame[i].wq);
	init_waitqueue_head(&spca50x->wq);

	/* setting availbe image modes */
	if (!spca50x_configure(spca50x)) {
		spca50x->user = 0;
		spca50x->last_times = 0;
		spca50x->dtimes = 0;
		spca50x->autoexpo = autoexpo;
		init_MUTEX(&spca50x->lock);	/* to 1 == available */
		init_MUTEX(&spca50x->buf_lock);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 10)
		spin_lock_init(&spca50x->v4l_lock);
#else
		spca50x->v4l_lock = SPIN_LOCK_UNLOCKED;
#endif
		spca50x->buf_state = BUF_NOT_ALLOCATED;
	} else {
		err("Failed to configure camera");
		goto error;
	}

	/* Init video stuff */
	spca50x->vdev = video_device_alloc();
	if (!spca50x->vdev)
		goto error;
	memcpy(spca50x->vdev, &spca50x_template, sizeof(spca50x_template));
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
	spca50x->vdev->dev = &dev->dev;
#endif
	video_set_drvdata(spca50x->vdev, spca50x);

	PDEBUG(2, "setting video device = %p, spca50x = %p", spca50x->vdev,
	       spca50x);

	if (video_register_device(spca50x->vdev, VFL_TYPE_GRABBER, video_nr) <
	    0) {
		err("video_register_device failed");
		goto error;
	}
	/* test on disconnect */
	spca50x->present = 1;
	/* Workaround for some applications that want data in RGB
	 * instead of BGR */
	if (spca50x->force_rgb)
		info("data format set to RGB");
#ifdef CONFIG_PROC_FS
	create_proc_spca50x_cam(spca50x);
#endif				/* CONFIG_PROC_FS */

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
	usb_set_intfdata(intf, spca50x);
	spca50x_create_sysfs(spca50x->vdev);
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	MOD_DEC_USE_COUNT;
	return spca50x;
#else
	return 0;
#endif

      error:
	if (spca50x->vdev) {
		if (spca50x->vdev->minor == -1)
			video_device_release(spca50x->vdev);
		else
			video_unregister_device(spca50x->vdev);
		spca50x->vdev = NULL;
	}
	if (spca50x) {
		kfree(spca50x);
		spca50x = NULL;
	}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	MOD_DEC_USE_COUNT;
	return NULL;
#else
	return -EIO;
#endif

      nodevice:
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	return NULL;
#else
	return -ENODEV;
#endif
}

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
static void spca5xx_disconnect(struct usb_interface *intf)
{
	struct usb_spca50x *spca50x = usb_get_intfdata(intf);
#else
static void spca5xx_disconnect(struct usb_device *dev, void *ptr)
{
	struct usb_spca50x *spca50x = (struct usb_spca50x *)ptr;
#endif
	int n;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	MOD_INC_USE_COUNT;
#endif
	if (!spca50x)
		return;
	down(&spca50x->lock);
	spca50x->present = 0;
	for (n = 0; n < SPCA50X_NUMFRAMES; n++)
		spca50x->frame[n].grabstate = FRAME_ABORTING;
	spca50x->curframe = -1;

	/* This will cause the process to request another frame */
	for (n = 0; n < SPCA50X_NUMFRAMES; n++)
		if (waitqueue_active(&spca50x->frame[n].wq))
			wake_up_interruptible(&spca50x->frame[n].wq);

	if (waitqueue_active(&spca50x->wq))
		wake_up_interruptible(&spca50x->wq);

	spca5xx_kill_isoc(spca50x);

	PDEBUG(3, "Disconnect Kill isoc done");
	up(&spca50x->lock);
	while (spca50x->user)
		schedule();
		
	{
		down(&spca50x->lock);
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
		/* be sure close did not use &intf->dev ? */
		dev_set_drvdata(&intf->dev, NULL);
#endif
		/* We don't want people trying to open up the device */
		if (spca50x->vdev)
			video_unregister_device(spca50x->vdev);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
		usb_driver_release_interface(&spca5xx_driver,
					     &spca50x->dev->actconfig->
					     interface[spca50x->iface]);
#endif
		spca50x->dev = NULL;
		up(&spca50x->lock);
#ifdef CONFIG_PROC_FS
		destroy_proc_spca50x_cam(spca50x);
#endif				/* CONFIG_PROC_FS */

		/* Free the memory */
		if (spca50x && !spca50x->user) {
			spca5xx_dealloc(spca50x);
			kfree(spca50x);
			spca50x = NULL;
		}
	}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	MOD_DEC_USE_COUNT;
#endif
	PDEBUG(3, "Disconnect complete");

}

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,22)
static struct usb_driver spca5xx_driver = {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
	.owner = THIS_MODULE,
#endif
	.name = "spca5xx",
	.id_table = device_table,
	.probe = spca5xx_probe,
	.disconnect = spca5xx_disconnect
};
#else
static struct usb_driver spca5xx_driver = {
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,20)
	THIS_MODULE,
#endif
	"spca5xx",
	spca5xx_probe,
	spca5xx_disconnect,
	{NULL, NULL}
};
#endif

/****************************************************************************
 *
 *  Module routines
 *
 ***************************************************************************/

static int __init usb_spca5xx_init(void)
{
#ifdef CONFIG_PROC_FS
	proc_spca50x_create();
#endif				/* CONFIG_PROC_FS */

	if (usb_register(&spca5xx_driver) < 0)
		return -1;

	info("spca5xx driver %s registered", version);

	return 0;
}

static void __exit usb_spca5xx_exit(void)
{
	usb_deregister(&spca5xx_driver);
	info("driver spca5xx deregistered");

#ifdef CONFIG_PROC_FS
	proc_spca50x_destroy();
#endif				/* CONFIG_PROC_FS */
}

module_init(usb_spca5xx_init);
module_exit(usb_spca5xx_exit);

//eof
